# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Tue Feb 11 12:29:35 2020
# Author: Zhi Huang, Purdue University
#
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

def expression_filter(x, meanq = 0.2, varq = 0.2):
    '''
    Parameters
    ----------
    x     : Real-valued expression matrix with rownames indicating
            gene ID or gene symbol.
    meanq : By which genes with low expression mean across samples are filtered out.
    varq  : By which genes with low expression variance across samples are filtered out.
        
    Returns
    -------
    x     : Real-valued expression matrix with rownames indicating
            gene ID or gene symbol.
    '''
    mean_quantile = x.mean(axis = 1).quantile(q=meanq)
    x = x.loc[x.mean(axis = 1) >= mean_quantile,:]
    var_quantile = x.var(axis = 1).quantile(q=varq)
    x = x.loc[x.var(axis = 1) >= var_quantile,:]
    return x



# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Mon Feb 10 17:57:08 2020
# Author: Zhi Huang, Purdue University
#      ___       ___           ___           ___           ___     
#     /\__\     /\__\         /\  \         /\  \         /\__\    
#    /:/  /    /::|  |       /::\  \       /::\  \       /::|  |   
#   /:/  /    /:|:|  |      /:/\:\  \     /:/\:\  \     /:|:|  |   
#  /:/  /    /:/|:|__|__    \:\~\:\  \   /:/  \:\  \   /:/|:|__|__ 
# /:/__/    /:/ |::::\__\    \:\ \:\__\ /:/__/ \:\__\ /:/ |::::\__\
# \:\  \    \/__/~~/:/  /     \:\/:/  / \:\  \  \/__/ \/__/~~/:/  /
#  \:\  \         /:/  /       \::/  /   \:\  \             /:/  / 
#   \:\  \       /:/  /        /:/  /     \:\  \           /:/  /  
#    \:\__\     /:/  /        /:/  /       \:\__\         /:/  /   
#     \/__/     \/__/         \/__/         \/__/         \/__/    
#
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

'''
Parameters
----------
    
data_in        : real-valued expression matrix with rownames indicating
                 gene ID or gene symbol.
gamma          : gamma value (default = 0.55)
t              : t value (default = 1)
lambda         : lambda value (default = 1)
beta           : beta value (default = 0.4)
minClusterSize : minimum length of cluster to retain (default = 10)
CCmethod       : Methods for correlation coefficient calculation (default =
                 "pearson"). Users can also pick "spearman".
normalization  : Determine if normalization is needed on massive correlation
                 coefficient matrix.
Returns
-------
None
Notes
-----
References
----------
.. [1] Zhang J, Huang K. Normalized lmqcm: An algorithm for detecting weak quasi-cliques
       in weighted graph with applications in gene co-expression module discovery in
       cancers. Cancer informatics. 2014 Jan;13:CIN-S14021.
.. [2] Huang Z, Han Z, Wang T, Shao W, Xiang S, Salama P, Rizkalla M, Huang K, Zhang J.
       TSUNAMI: Translational Bioinformatics Tool Suite For Network Analysis And Mining.
       bioRxiv. 2019 Jan 1:787507.
Examples
-------
>>> tcga_COAD_data = 'http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/COAD/20160128/gdac.broadinstitute.org_COAD.Merge_rnaseqv2__illuminahiseq_rnaseqv2__unc_edu__Level_3__RSEM_genes_normalized__data.Level_3.2016012800.0.0.tar.gz'
>>> data_in = pd.read_csv(tcga_COAD_data, header=0, skiprows=range(1, 2), index_col=0, sep='\t')
>>> lobject = lmQCM(data_in)
>>> lobject.fit()
>>> lobject.clusters
>>> lobject.clusters_names
>>> lobject.eigengene_matrix
'''

import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.stats import spearmanr


class lmQCM():
    def __init__(self, data_in = None, gamma = 0.55, t = 1, lambdaa = 1, beta = 0.4, 
                     minClusterSize = 10, CCmethod = "pearson", normalization = False):
        
        self.data_in = data_in
        if 'DataFrame' not in str(type(self.data_in)):
            print('Input matrix is numpy matrix. Convert it to pandas.core.frame.DataFrame...')
            self.data_in = pd.DataFrame(self.data_in)
        if np.sum(np.isnan(self.data_in.values)) > 0:
            print('Warning: %d NaN value detected. Replacing them to zero...' % np.sum(np.isnan(self.data_in.values)))
            self.data_in.fillna(0, inplace = True)
        self.gamma = gamma
        self.t = t
        self.lambdaa = lambdaa
        self.beta = beta
        self.minClusterSize = minClusterSize
        self.CCmethod = CCmethod
        self.normalization = normalization
        self.calculate_correlation_matrix()
        print('Initialization Done.')
    
    def localMaximumQCM(self):
        C = []
        nRow = self.corr_mat.shape[0]
        maxV = np.max(self.corr_mat, axis = 0)
        maxInd = np.argmax(self.corr_mat, axis = 1)
        lm_ind = np.where(maxV == np.max(self.corr_mat[maxInd,], axis = 1))[0]
        maxEdges = np.stack((maxInd[lm_ind], lm_ind)).T
        maxW = maxV[lm_ind]
        sortMaxV = np.sort(maxW, kind='mergesort')[::-1] # decreasing
        sortMaxInd = np.argsort(maxW, kind='mergesort')[::-1]
        sortMaxEdges = maxEdges[sortMaxInd, ]
        print("Number of Maximum Edges: %d" % len(sortMaxInd))
        currentInit = 1
        noNewInit = 0
        
        pbar = tqdm(total=len(sortMaxInd))
        nodesInCluster = []
        while currentInit <= len(sortMaxInd) and noNewInit == 0:
            pbar.update(1)
            if sortMaxV[currentInit] < (self.gamma * sortMaxV[1]):
                noNewInit = 1
            else:
                if sortMaxEdges[currentInit, 0] not in nodesInCluster and sortMaxEdges[currentInit, 1] not in nodesInCluster:
                    newCluster = list(sortMaxEdges[currentInit, ])
                    addingMode = 1
                    currentDensity = sortMaxV[currentInit]
                    nCp = 2
                    totalInd = np.arange(nRow)
                    remainInd = np.setdiff1d(totalInd, newCluster)
                    while addingMode == 1:
                        neighborWeights = np.sum(self.corr_mat[newCluster,:][:,remainInd], axis = 0)
                        maxNeighborWeight = max(neighborWeights)
                        maxNeighborInd = np.argmax(neighborWeights)
                        c_v = maxNeighborWeight/nCp
                        alphaN = 1 - 1/(2 * self.lambdaa * (nCp + self.t))
                        if c_v >= alphaN * currentDensity:
                            newCluster = newCluster + [remainInd[maxNeighborInd]]
                            nCp = nCp + 1
                            currentDensity = (currentDensity * ((nCp - 1) * (nCp - 2)/2) + maxNeighborWeight)/(nCp * (nCp - 1)/2)
                            remainInd = np.setdiff1d(remainInd, remainInd[maxNeighborInd])
                        else:
                            addingMode = 0
                    nodesInCluster = nodesInCluster + newCluster
                    C = C + [newCluster]
            currentInit += 1
        print(" Calculation Finished.")
        pbar.close()
        return(C)
        
    def merging_lmQCM(self, C):
        sizeC = [len(i) for i in C]
        sortInd = np.argsort(sizeC, kind='mergesort')[::-1]
        mergedCluster = [C[i] for i in sortInd if len(C[i]) >= self.minClusterSize]
        mergeOccur = 1
        currentInd = -1
        print(" %d Modules before merging." % len(mergedCluster))
        while mergeOccur == 1:
            mergeOccur = 0
            while currentInd < len(mergedCluster):
                currentInd += 1
                if currentInd < len(mergedCluster):
                    keepInd = list(np.arange(0,currentInd+1))
                    for j in np.arange(currentInd+1, len(mergedCluster)):
                        interCluster = np.intersect1d(mergedCluster[currentInd], mergedCluster[j])
                        if len(interCluster) >= self.beta * min(len(mergedCluster[j]), len(mergedCluster[currentInd])):
                            mergedCluster[currentInd] = list(np.union1d(mergedCluster[currentInd], mergedCluster[j]))
                            mergeOccur = 1
                        else:
                            keepInd += [j]
                    mergedCluster = [mergedCluster[i] for i in keepInd]
                    
            sizeMergedCluster = [len(mergedCluster[i]) for i in range(len(mergedCluster))]
            sortMergedInd = np.argsort(sizeMergedCluster, kind='mergesort')[::-1]
            mergedCluster = [mergedCluster[i] for i in sortMergedInd]
            currentInd = 0
        print(" %d Modules remain after merging." % len(mergedCluster))
        return mergedCluster
        
    def calculate_correlation_matrix(self):
        print("Calculating massive correlation coefficient ...")
        if self.CCmethod.lower() == "pearson": self.corr_mat = np.corrcoef(self.data_in.values)
        if self.CCmethod.lower() == "spearman": self.corr_mat = spearmanr(self.data_in.values.T).correlation
        np.fill_diagonal(self.corr_mat, 0)
        if np.sum(np.isnan(self.corr_mat)) > 0:
            print('Warning: %d NaN value detected in correlation matrix. Replacing them to zero...' % np.sum(np.isnan(self.corr_mat)))
            self.corr_mat[np.isnan(self.corr_mat)] = 0
        if self.normalization:
            self.corr_mat = np.abs(self.corr_mat)
            D = np.sum(self.corr_mat, axis = 0)
            D_half = 1.0/np.sqrt(D)
            self.corr_mat = np.multiply(np.multiply(self.corr_mat, D_half).T, D_half)
        
    def fit(self):
        C = self.localMaximumQCM()
        clusters = self.merging_lmQCM(C)
        
        clusters_names = []
        for i in range(len(clusters)):
            mc = clusters[i]
            clusters_names.append(list(self.data_in.index.values[mc]))
        eigengene_matrix = np.zeros((len(clusters), self.data_in.shape[1]))
        for i in range(len(clusters_names)):
            geneID = clusters_names[i]
            X = self.data_in.loc[geneID, ]
            mu = np.nanmean(X, axis = 1) # rowMeans
            stddev = np.nanstd(X, axis = 1, ddof= 1) # ddof=1 provides unbiased estimation (1/(n-1))
            XNorm = (X.T-mu).T
            XNorm = (XNorm.T/stddev).T
            u, s, vh = np.linalg.svd(XNorm, full_matrices = False)
            eigengene_matrix[i, ] = vh[0,:]
        eigengene_matrix = pd.DataFrame(eigengene_matrix, columns = self.data_in.columns)
        self.clusters = clusters
        self.clusters_names = clusters_names
        self.eigengene_matrix = eigengene_matrix
        return self.clusters, self.clusters_names, self.eigengene_matrix
    
    
    
    


# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Tue Feb 12 16:00:51 2020
# Author: Zhi Huang, Purdue University
#
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#
import numpy as np
import matplotlib.pyplot as plt
from lifelines.statistics import logrank_test
from lifelines.utils import concordance_index
from lifelines import KaplanMeierFitter

def logranktest(hazard_ratio, time, event):
    groups = (hazard_ratio > np.median(hazard_ratio)).astype(int)
    # 0: low hazard ratio
    # 1: high hazard ratio
    
    logrank_results = logrank_test(time[groups == 0], time[groups == 1], event_observed_A=event[groups == 0], event_observed_B=event[groups == 1])
    p_value = logrank_results.p_value
#    test_statistic = logrank_results.test_statistic
    # plots
    kmf = KaplanMeierFitter()
    fig = plt.figure()
    ax = plt.subplot(111)
    kmf.fit(time[groups == 0], event_observed=event[groups == 0], label='Group 1')
    kmf.plot(ax=ax, show_censors = True, ci_show = False, color = 'black')
    kmf.fit(time[groups == 1], event_observed=event[groups == 1], label='Group 2')
    kmf.plot(ax=ax, show_censors = True, ci_show = False, color = 'red')
    ax.text(0, ax.get_ylim()[0]+0.05, s = '  p-value = %.4e' % p_value, fontsize = 14)
    ax.grid(color='grey', linestyle='--', linewidth=0.5)
    ax.set_xlabel('Time')
    ax.set_ylabel('Survival probability')
    ax.set_title('Survival Plot', fontsize = 14)
    
    return logrank_results, fig



# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Tue Feb 11 12:29:35 2020
# Author: Zhi Huang, Purdue University
#
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#
import numpy as np

def fisher_discriminant(H, label):
    '''
    Parameters
    ----------
    H     : Real-valued matrix with columns indicating samples.
    label : Class indices.
        
    Returns
    -------
    E_D   : Real scalar value indicating fisher discriminant.
    Notes
    -----
    This fisher discriminant is the equation (3 a,b) in
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3495075
    
    label is further sorted in ascending order, then apply its order to label and H.
    Otherwise denominator will be wrong.
    
    References
    ----------
    .. [1] Lee SY, Song HA, Amari SI. A new discriminant NMF algorithm and its
           application to the extraction of subtle emotional differences in speech.
           Cognitive neurodynamics. 2012 Dec 1;6(6):525-35.

    '''
    order = np.argsort(label)
    H = H[:,order]
    label = label[order]
    numerator, denominator = 0, 0
    mu_rkn = np.zeros((H.shape[0], 0))
    mu_r_all = 1/H.shape[1] * np.sum(H, axis = 1)
    for k in np.unique(label):
        N_k = np.sum(k == label)
        mu_rk_block = np.zeros((0, N_k))
        for r in range(H.shape[0]):
            mu_r = mu_r_all[r]
            mu_rk = 1/N_k * np.sum(H[r, k == label])
            mu_rk_block = np.concatenate((mu_rk_block, np.array([mu_rk] * N_k).reshape(1,N_k)), axis = 0)
            numerator += N_k * (mu_rk - mu_r) ** 2
        mu_rkn = np.concatenate((mu_rkn, mu_rk_block), axis = 1)
    denominator = np.sum((H - mu_rkn)**2)
    E_D = numerator / denominator
    return E_D


# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Wed Feb 19 14:23:17 2020
# Author: Zhi Huang, Purdue University
#
# This is a Child Class inherited from lifelines.CoxPHFitter
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#
import numpy as np
from numpy import dot, einsum, log, exp, zeros, arange, multiply, ndarray
import pandas as pd
from pandas import DataFrame, Series, Index
from lifelines.utils import normalize, coalesce, CensoringType
from lifelines import CoxPHFitter
from numpy.linalg import inv
from datetime import datetime
from typing import Callable, Iterator, List, Optional, Tuple, Union, Any, Iterable
import warnings

class StepCoxPHFitter(CoxPHFitter):
    def __init__(self, tie_method="Efron", penalizer=0.0, l1_ratio = 0.0, strata=None, baseline_estimation_method="breslow"):
        super(CoxPHFitter, self).__init__()
        if penalizer < 0:
            raise ValueError("penalizer parameter must be >= 0.")
        if tie_method != "Efron":
            raise NotImplementedError("Only Efron is available at the moment.")
        self.baseline_estimation_method = baseline_estimation_method
        self.l1_ratio = l1_ratio
        self.tie_method = tie_method
        self.penalizer = penalizer
        self.strata = strata
        self.cph_max_steps = 50
 

    @CensoringType.right_censoring
    def fit(
        self,
        df,
        duration_col=None,
        event_col=None,
        show_progress=False,
        initial_point=None,
        strata=None,
        step_size=None,
        weights_col=None,
        cluster_col=None,
        robust=False,
        batch_mode=None,
    ):
        """
        Fit the Cox proportional hazard model to a dataset.
        Parameters
        ----------
        df: DataFrame
            a Pandas DataFrame with necessary columns `duration_col` and
            `event_col` (see below), covariates columns, and special columns (weights, strata).
            `duration_col` refers to
            the lifetimes of the subjects. `event_col` refers to whether
            the 'death' events was observed: 1 if observed, 0 else (censored).
        duration_col: string
            the name of the column in DataFrame that contains the subjects'
            lifetimes.
        event_col: string, optional
            the  name of thecolumn in DataFrame that contains the subjects' death
            observation. If left as None, assume all individuals are uncensored.
        weights_col: string, optional
            an optional column in the DataFrame, df, that denotes the weight per subject.
            This column is expelled and not used as a covariate, but as a weight in the
            final regression. Default weight is 1.
            This can be used for case-weights. For example, a weight of 2 means there were two subjects with
            identical observations.
            This can be used for sampling weights. In that case, use `robust=True` to get more accurate standard errors.
        show_progress: boolean, optional (default=False)
            since the fitter is iterative, show convergence
            diagnostics. Useful if convergence is failing.
        initial_point: (d,) numpy array, optional
            initialize the starting point of the iterative
            algorithm. Default is the zero vector.
        strata: list or string, optional
            specify a column or list of columns n to use in stratification. This is useful if a
            categorical covariate does not obey the proportional hazard assumption. This
            is used similar to the `strata` expression in R.
            See http://courses.washington.edu/b515/l17.pdf.
        step_size: float, optional
            set an initial step size for the fitting algorithm. Setting to 1.0 may improve performance, but could also hurt convergence.
        robust: boolean, optional (default=False)
            Compute the robust errors using the Huber sandwich estimator, aka Wei-Lin estimate. This does not handle
            ties, so if there are high number of ties, results may significantly differ. See
            "The Robust Inference for the Cox Proportional Hazards Model", Journal of the American Statistical Association, Vol. 84, No. 408 (Dec., 1989), pp. 1074- 1078
        cluster_col: string, optional
            specifies what column has unique identifiers for clustering covariances. Using this forces the sandwich estimator (robust variance estimator) to
            be used.
        batch_mode: bool, optional
            enabling batch_mode can be faster for datasets with a large number of ties. If left as None, lifelines will choose the best option.
        Returns
        -------
        self: CoxPHFitter
            self with additional new properties: ``print_summary``, ``hazards_``, ``confidence_intervals_``, ``baseline_survival_``, etc.
        Note
        ----
        Tied survival times are handled using Efron's tie-method.
        Examples
        --------
        >>> from lifelines import CoxPHFitter
        >>>
        >>> df = pd.DataFrame({
        >>>     'T': [5, 3, 9, 8, 7, 4, 4, 3, 2, 5, 6, 7],
        >>>     'E': [1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0],
        >>>     'var': [0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2],
        >>>     'age': [4, 3, 9, 8, 7, 4, 4, 3, 2, 5, 6, 7],
        >>> })
        >>>
        >>> cph = CoxPHFitter()
        >>> cph.fit(df, 'T', 'E')
        >>> cph.print_summary()
        >>> cph.predict_median(df)
        >>> from lifelines import CoxPHFitter
        >>>
        >>> df = pd.DataFrame({
        >>>     'T': [5, 3, 9, 8, 7, 4, 4, 3, 2, 5, 6, 7],
        >>>     'E': [1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0],
        >>>     'var': [0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2],
        >>>     'weights': [1.1, 0.5, 2.0, 1.6, 1.2, 4.3, 1.4, 4.5, 3.0, 3.2, 0.4, 6.2],
        >>>     'month': [10, 3, 9, 8, 7, 4, 4, 3, 2, 5, 6, 7],
        >>>     'age': [4, 3, 9, 8, 7, 4, 4, 3, 2, 5, 6, 7],
        >>> })
        >>>
        >>> cph = CoxPHFitter()
        >>> cph.fit(df, 'T', 'E', strata=['month', 'age'], robust=True, weights_col='weights')
        >>> cph.print_summary()
        >>> cph.predict_median(df)
        """
        if duration_col is None:
            raise TypeError("duration_col cannot be None.")

        self._time_fit_was_called = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S") + " UTC"
        self.duration_col = duration_col
        self.event_col = event_col
        self.robust = robust
        self.cluster_col = cluster_col
        self.weights_col = weights_col
        self._n_examples = df.shape[0]
        self._batch_mode = batch_mode
        self.strata = coalesce(strata, self.strata)

        X, T, E, weights, original_index, self._clusters = self._preprocess_dataframe(df)

        self.durations = T.copy()
        self.event_observed = E.copy()
        self.weights = weights.copy()

        if self.strata is not None:
            self.durations.index = original_index
            self.event_observed.index = original_index
            self.weights.index = original_index

        self._norm_mean = X.mean(0)
        self._norm_std = X.std(0)

        # this is surprisingly faster to do...
        X_norm = pd.DataFrame(
            normalize(X.values, self._norm_mean.values, self._norm_std.values), index=X.index, columns=X.columns
        )

        params_, ll_, variance_matrix_, baseline_hazard_, baseline_cumulative_hazard_ = self._fit_model(
            X_norm, T, E, weights=weights, initial_point=initial_point, show_progress=show_progress, step_size=step_size, max_steps = self.cph_max_steps
        )

        self.log_likelihood_ = ll_
        self.variance_matrix_ = variance_matrix_
        self.params_ = pd.Series(params_, index=X.columns, name="coef")
        self.baseline_hazard_ = baseline_hazard_
        self.baseline_cumulative_hazard_ = baseline_cumulative_hazard_

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._predicted_partial_hazards_ = (
                self.predict_partial_hazard(X)
                .to_frame(name="P")
                .assign(T=self.durations.values, E=self.event_observed.values, W=self.weights.values)
                .set_index(X.index)
            )

        self.standard_errors_ = self._compute_standard_errors(X_norm, T, E, weights)
        self.confidence_intervals_ = self._compute_confidence_intervals()
        self.baseline_survival_ = self._compute_baseline_survival()

        if hasattr(self, "_concordance_index_"):
            del self._concordance_index_

        return self
    
    
    
    def _fit_model_breslow(
        self,
        X: DataFrame,
        T: Series,
        E: Series,
        weights: Series,
        initial_point: Optional[ndarray] = None,
        step_size: Optional[float] = None,
        show_progress: bool = True,
        max_steps: int = 1
    ):
        beta_, ll_, hessian_ = self._newton_rhapson_for_efron_model(
            X, T, E, weights, initial_point=initial_point, step_size=step_size, show_progress=show_progress
        )

        # compute the baseline hazard here.
        predicted_partial_hazards_ = (
            pd.DataFrame(np.exp(dot(X, beta_)), columns=["P"])
            .assign(T=T.values, E=E.values, W=weights.values)
            .set_index(X.index)
        )
        baseline_hazard_ = self._compute_baseline_hazards(predicted_partial_hazards_)
        baseline_cumulative_hazard_ = self._compute_baseline_cumulative_hazard(baseline_hazard_)

        # rescale parameters back to original scale.
        params_ = beta_ / self._norm_std.values
        if hessian_.size > 0:
            variance_matrix_ = pd.DataFrame(
                -inv(hessian_) / np.outer(self._norm_std, self._norm_std), index=X.columns, columns=X.columns
            )
        else:
            variance_matrix_ = pd.DataFrame(index=X.columns, columns=X.columns)

        return params_, ll_, variance_matrix_, baseline_hazard_, baseline_cumulative_hazard_



# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Wed Feb 19 13:20:25 2020
# Author: Zhi Huang, Purdue University
#
# This is a concise version rewrite from sklearn_decomposition_nmf.
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#
import numpy as np
import scipy.sparse as sp
import pandas as pd
from sklearn.utils import check_random_state, check_array
from sklearn.decomposition.nmf import _initialize_nmf
from sklearn.decomposition.cdnmf_fast import _update_cdnmf_fast
from sklearn.utils.extmath import safe_sparse_dot
import copy

from lifelines.utils import concordance_index
from ..survival import StepCoxPHFitter
import time
import warnings
import logging
EPSILON = np.finfo(np.float32).eps

    
def trace_dot(X, Y):
    """Trace of np.dot(X, Y.T).

    Parameters
    ----------
    X : array-like
        First matrix
    Y : array-like
        Second matrix
    """
    return np.dot(X.ravel(), Y.ravel())

def squared_norm(x):
    """Squared Euclidean or Frobenius norm of x.
    Faster than norm(x) ** 2.
    Parameters
    ----------
    x : array_like
    Returns
    -------
    float
        The Euclidean norm when x is a vector, the Frobenius norm when x
        is a matrix (2-d array).
    """
    x = np.ravel(x, order='K')
    if np.issubdtype(x.dtype, np.integer):
        warnings.warn('Array type is integer, np.dot may overflow. '
                      'Data should be float type to avoid this issue',
                      UserWarning)
    return np.dot(x, x)

def calcuate_Frobenius_norm(X, W, H, square_root=False):
    """Compute the beta-divergence of X and dot(W, H).

    Parameters
    ----------
    X : float or array-like, shape (n_samples, n_features)

    W : float or dense array-like, shape (n_samples, n_components)

    H : float or dense array-like, shape (n_components, n_features)

    Returns
    -------
        res : float
            Frobenius norm of X and np.dot(W, H)
    """

    # The method can be called with scalars
    if not sp.issparse(X):
        X = np.atleast_2d(X)
    W = np.atleast_2d(W)
    H = np.atleast_2d(H)

    # Frobenius norm
    # Avoid the creation of the dense np.dot(W, H) if X is sparse.
    if sp.issparse(X):
        norm_X = np.dot(X.data, X.data)
        norm_WH = trace_dot(np.dot(np.dot(W.T, W), H), H)
        cross_prod = trace_dot((X * H.T), W)
        res = (norm_X + norm_WH - 2. * cross_prod) / 2.
    else:
        res = squared_norm(X - np.dot(W, H)) / 2.

    if square_root:
        return np.sqrt(res * 2)
    else:
        return res

def _multiplicative_update_w(X, W, H, HHt=None, XHt=None, update_H=True):
    """update W in Multiplicative Update NMF"""
    # Numerator
    if XHt is None:
        XHt = safe_sparse_dot(X, H.T)
    if update_H:
        # avoid a copy of XHt, which will be re-computed (update_H=True)
        numerator = XHt
    else:
        # preserve the XHt, which is not re-computed (update_H=False)
        numerator = XHt.copy()

    # Denominator
    if HHt is None:
        HHt = np.dot(H, H.T)
    denominator = np.dot(W, HHt)
    denominator[denominator == 0] = EPSILON
    numerator /= denominator
    delta_W = numerator
    return delta_W, HHt, XHt

def _multiplicative_update_w_orth(X, W, H, HHt=None, XHt=None, sigma=0):
    '''
        Implemented based on equation (18) from:
        Mirzal, Andri. "A convergent algorithm for orthogonal nonnegative matrix factorization."
        Journal of Computational and Applied Mathematics 260 (2014): 149-166.
    '''
    if XHt is None:
        XHt = safe_sparse_dot(X, H.T)
        
    numerator = XHt + sigma*W
    # Denominator
    if HHt is None:
        HHt = np.dot(H, H.T)
    # ONMF on W
    denominator = np.dot(W, HHt) + sigma * W.dot(W.T).dot(W)
    denominator[denominator == 0] = EPSILON
    numerator /= denominator
    delta_W = numerator
    
#    # ONMF on W
#    denominator = W.dot(W.T).dot(X).dot(H.T) # Ding et al. (2006) Orthogonal Nonnegative Matrix Tri-factorizations for Clustering
#    delta_W = np.sqrt(numerator)
    
    return delta_W, HHt, XHt


def _multiplicative_update_h(X, W, H, beta_loss, l1_reg_H, l2_reg_H, gamma):
    """update H in Multiplicative Update NMF"""
    numerator = safe_sparse_dot(W.T, X)
    denominator = np.dot(np.dot(W.T, W), H)
    denominator[denominator == 0] = EPSILON
    numerator /= denominator
    delta_H = numerator
    return delta_H

def _update_coordinate_descent(X, W, Ht, shuffle,
                               random_state):
    """Helper function for _fit_coordinate_descent

    Update W to minimize the objective function, iterating once over all
    coordinates. By symmetry, to update H, one can call
    _update_coordinate_descent(X.T, Ht, W, ...)

    """
    n_components = Ht.shape[1]

    HHt = np.dot(Ht.T, Ht)
    XHt = safe_sparse_dot(X, Ht)

    if shuffle:
        permutation = random_state.permutation(n_components)
    else:
        permutation = np.arange(n_components)
    # The following seems to be required on 64-bit Windows w/ Python 3.5.
    permutation = np.asarray(permutation, dtype=np.intp)
    return _update_cdnmf_fast(W, HHt, XHt, permutation)



def NMF(X, n_components, solver = 'cd', max_iter=1000, tol=1e-6, update_H = True, random_state=None, shuffle=False, verbose=0):
    '''
    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Constant input matrix.

    W : array-like, shape (n_samples, n_components)
        Initial guess for the solution.

    H : array-like, shape (n_components, n_features)
        Initial guess for the solution.
    '''
    W, H = _initialize_nmf(X, n_components, init = 'random', random_state=random_state)
    
    if solver == 'mu':
        # used for the convergence criterion
        error_at_init = calcuate_Frobenius_norm(X, W, H, square_root=True)
        previous_error = error_at_init
        
        start_time = time.time()
        HHt, XHt = None, None
        for n_iter in range(1, max_iter + 1):
            # update W
            # HHt and XHt are saved and reused if not update_H
            delta_W, HHt, XHt = _multiplicative_update_w(X, W, H, HHt, XHt, update_H = update_H)
            W *= delta_W
            
            # update H
            if update_H:
                delta_H = _multiplicative_update_h(X, W, H)
                H *= delta_H
                # These values will be recomputed since H changed
                HHt, XHt = None, None
        
            # test convergence criterion every 10 iterations
            if tol > 0 and n_iter % 10 == 0:
                error = calcuate_Frobenius_norm(X, W, H, square_root=True)
                if verbose:
                    iter_time = time.time()
                    print("Epoch %02d reached after %.3f seconds, error: %f" %
                          (n_iter, iter_time - start_time, error))
                if (previous_error - error) / error_at_init < tol:
                    break
                previous_error = error
    
        # do not print if we have already printed in the convergence test
        if verbose and (tol == 0 or n_iter % 10 != 0):
            end_time = time.time()
            print("Epoch %02d reached after %.3f seconds." %
                  (n_iter, end_time - start_time))
        return W, H, n_iter
            
    if solver == 'cd':
        # so W and Ht are both in C order in memory
        Ht = check_array(H.T, order='C')
        X = check_array(X, accept_sparse='csr')
    
        rng = check_random_state(random_state)
    
        for n_iter in range(max_iter):
            violation = 0.
    
            # Update W
            violation += _update_coordinate_descent(X, W, Ht, shuffle, rng)
            # Update H
            if update_H:
                violation += _update_coordinate_descent(X.T, Ht, W, shuffle, rng)
    
            if n_iter == 0:
                violation_init = violation
    
            if violation_init == 0:
                break
    
            if verbose:
                print("violation:", violation / violation_init)
    
            if violation / violation_init <= tol:
                if verbose:
                    print("Converged at iteration", n_iter + 1)
                break
            
        return W, Ht.T, n_iter
        

def CoxNMF(X, t, e, n_components, alpha=1e-5, sigma = 0, eta_b = None, cph_penalizer=0, l1_ratio = 0, ci_tol=0.02, solver='mu', update_rule='projection', cph_max_steps=1, max_iter=1000, tol=1e-6, random_state=None, update_H=True, update_beta=True, H_row_normalization=False, logger=None, verbose=0):
    '''
    Parameters
    ----------
    X     : array-like, shape (n_samples, n_features)
            Constant input matrix.

    W     : array-like, shape (n_samples, n_components)
            Initial guess for the solution.

    H     : array-like, shape (n_components, n_features)
            Initial guess for the solution.
        
    t     : array-like, shape (n_components)
            Survival time.
        
    e     : array-like, shape (n_components)
            Survival event (death = 1).
        
    alpha : scalar value.
            parameter used for learning the H guided by Cox model.
            
    sigma : scalar value.
            orthogonal constraint on W.
            
    eta_b : scalar value.
            step size in Cox model.
            
    ci_tol: Tolerace of decrease of oncordance index to stop iteration.
    '''
    W, H = _initialize_nmf(X, n_components, init = 'random', random_state=random_state)
        
    # used for the convergence criterion
    error_at_init = calcuate_Frobenius_norm(X, W, H, square_root=True)
    previous_error = error_at_init
    max_cindex = 0.5
    start_time = time.time()
    HHt, XHt = None, None
    t_geq_matrix = np.array([[int(y >= x) for i,x in enumerate(t)] for j,y in enumerate(t)])
    error_list = []
    cindex_list = []
    for n_iter in range(1, max_iter + 1):
        # update W
        # HHt and XHt are saved and reused if not update_H
        if sigma == 0:
            delta_W, HHt, XHt = _multiplicative_update_w(X, W, H, HHt, XHt, update_H=update_H)
        elif sigma > 0:
            delta_W, HHt, XHt = _multiplicative_update_w_orth(X, W, H, HHt, XHt, sigma = sigma)
            
        W *= delta_W
#        if H_row_normalization:
#            W = (W.T / np.linalg.norm(W, axis=1).T).T
        
        beta, cph = None, None
        if update_beta:
            H_cox = pd.DataFrame(H.T)
            H_cox['time'] = t
            H_cox['event'] = e
            if cph_penalizer > 0:
                cph = StepCoxPHFitter(penalizer = cph_penalizer, l1_ratio = l1_ratio)
            else:
                cph = StepCoxPHFitter()
            cph.max_iterations = cph_max_steps
            if not beta:
                initial_point = None
            else:
                initial_point = beta.reshape(-1)
            cph.fit(H_cox, initial_point = initial_point, duration_col='time', event_col='event', step_size = eta_b, show_progress=False)
            cindex = concordance_index(H_cox['time'], -cph.predict_partial_hazard(H.T), H_cox['event'])
            beta = cph.params_.values.reshape(-1,1)
        # update H
        if update_H:
            n_patients = t.shape[0]
            numerator = safe_sparse_dot(W.T, X)
            denominator = np.dot(np.dot(W.T, W), H)
            H_mu = H*(numerator/denominator)
            
            cox_numerator = np.repeat(np.expand_dims(np.matmul(beta, np.exp(np.matmul(beta.T, H)) ), axis = 2), n_patients, axis = 2).swapaxes(1,2) * t_geq_matrix.T
            cox_numerator[:, np.where(e==0)[0], :] = 0
            cox_denominator = np.expand_dims(np.matmul(np.exp(np.matmul(beta.T, H)), t_geq_matrix), axis = 2)
            cox_fraction = e * np.repeat(beta, n_patients, axis = 1) - np.sum(cox_numerator / cox_denominator, axis = 1)

            H_partial = alpha / 2 * (numerator/denominator) * cox_fraction
            
            if update_rule == 'projection':
                H_partial[H_partial < 0] = 0
            
            H = H_mu + H_partial
            if H_row_normalization:
                H = (H.T / np.linalg.norm(H, axis=1)).T
            
            # These values will be recomputed since H changed
            HHt, XHt = None, None
    
        error = calcuate_Frobenius_norm(X, W, H, square_root=True)
        if verbose:
            print("Epoch %04d error: %f, concordance index: %f" % (n_iter, error, cindex))
        if logger:
            logger.log(logging.INFO, "Epoch %04d error: %f, concordance index: %f" % (n_iter, error, cindex))
        
        error_list.append(error)
        cindex_list.append(cindex)
        
        # test convergence criterion every 10 iterations
#        if tol > 0 and n_iter % 10 == 0:
        if (previous_error - error) / error_at_init < tol:
            break
        previous_error = error
        if (cindex - max_cindex) < -ci_tol: # if new concordance index smaller than previous 0.02
            break
        
        if cindex >= max_cindex:
            max_cindex = cindex
            max_cindex_res = {}
            max_cindex_res['W'] = W
            max_cindex_res['H'] = H
            max_cindex_res['cph'] = copy.deepcopy(cph)
            max_cindex_res['error'] = error
            max_cindex_res['cindex'] = cindex
            
        
    # do not print if we have already printed in the convergence test
    if verbose and (tol == 0 or n_iter % 10 != 0):
        end_time = time.time()
        print("Epoch %04d reached after %.3f seconds." %
              (n_iter, end_time - start_time))
    return W, H, cph, n_iter, error_list, cindex_list, max_cindex_res







# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Mon Feb 10 17:57:08 2020
# Author: Zhi Huang, Purdue University
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="biolearns",
    version="0.0.44",
    author="Zhi Huang",
    author_email="huang898@purdue.edu",
    description="BioLearns: Computational Biology and Bioinformatics Toolbox in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://biolearns.com",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



# Copyright 2020 Zhi Huang.  All rights reserved
# Created on Tue Feb 11 12:40:52 2020
# Author: Zhi Huang, Purdue University
#
#
# The original code came with the following disclaimer:
#
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fitness for a given application.  In no event
# shall Zhi Huang be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import pandas as pd
import numpy as np
import requests
import sys
import os
import re
import tarfile
import tempfile
'''
Parameters
----------
None
    
Returns
-------
None

Notes
-----
Data is downloaded directly from https://gdac.broadinstitute.org/.
The results here are in whole or part based upon data generated by 
the TCGA Research Network: https://www.cancer.gov/tcga.

References
----------
None
       
Examples
--------
>>> brca = TCGA('BRCA')
>>> brca.get_mRNAseq()
>>> brca.get_clinical()

'''

class TCGA():
    def __init__(self, cohort = 'ACC', download = True):
        self.cohort = cohort.upper()
        self.cohortdict = { 'ACC':'Adrenocortical carcinoma',
                            'BLCA':'Bladder urothelial carcinoma',
                            'BRCA':'Breast invasive carcinoma',
                            'CESC':'Cervical and endocervical cancers',
                            'CHOL':'Cholangiocarcinoma',
                            'COAD':'Colon adenocarcinoma',
                            'COADREAD':'Colorectal adenocarcinoma',
                            'DLBC':'Lymphoid Neoplasm Diffuse Large B-cell Lymphoma',
                            'ESCA':'Esophageal carcinoma',
                            'FPPP':'FFPE Pilot Phase II',
                            'GBM':'Glioblastoma multiforme',
                            'GBMLGG':'Glioma',
                            'HNSC':'Head and Neck squamous cell carcinoma',
                            'KICH':'Kidney Chromophobe',
                            'KIPAN':'Pan-kidney cohort (KICH+KIRC+KIRP)',
                            'KIRC':'Kidney renal clear cell carcinoma',
                            'KIRP':'Kidney renal papillary cell carcinoma',
                            'LAML':'Acute Myeloid Leukemia',
                            'LGG':'Brain Lower Grade Glioma',
                            'LIHC':'Liver hepatocellular carcinoma',
                            'LUAD':'Lung adenocarcinoma',
                            'LUSC':'Lung squamous cell carcinoma',
                            'MESO':'Mesothelioma',
                            'OV':'Ovarian serous cystadenocarcinoma',
                            'PAAD':'Pancreatic adenocarcinoma',
                            'PCPG':'Pheochromocytoma and Paraganglioma',
                            'PRAD':'Prostate adenocarcinoma',
                            'READ':'Rectum adenocarcinoma',
                            'SARC':'Sarcoma',
                            'SKCM':'Skin Cutaneous Melanoma',
                            'STAD':'Stomach adenocarcinoma',
                            'STES':'Stomach and Esophageal carcinoma',
                            'TGCT':'Testicular Germ Cell Tumors',
                            'THCA':'Thyroid carcinoma',
                            'THYM':'Thymoma',
                            'UCEC':'Uterine Corpus Endometrial Carcinoma',
                            'UCS':'Uterine Carcinosarcoma',
                            'UVM':'Uveal Melanoma'}
        if download:
#            self.tmpdir = tempfile.TemporaryDirectory()
            self.mRNAseq = self.get_mRNAseq()
            self.miRSeq = None
            self.mRNA = None
            self.RPPA = None
            self.methylation = None
            self.clinical = self.get_clinical()
            self.overall_survival_time, self.overall_survival_event = self._get_overall_survival(death_censor = True)
        
    def get_mRNAseq(self):
        print('Retrieve mRNAseq from http://firebrowse.org/ ...')
        print('Cohort: %s (%s)' % (self.cohortdict[self.cohort], self.cohort))
        print('File type: illuminahiseq_rnaseqv2-RSEM_genes_normalized')

        link = 'https://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/' + self.cohort + \
                '/20160128/gdac.broadinstitute.org_' + self.cohort + \
                '.Merge_rnaseqv2__illuminahiseq_rnaseqv2__unc_edu__Level_3__RSEM_genes_normalized__data.Level_3.2016012800.0.0.tar.gz'

        if self.cohort in ['FPPP']:
            print('No mRNAseq data.')
            return
            
        file_downloaded = '/tmp/' + self.cohort + '.mRNAseq.tar.gz'
        
        if not os.path.exists(file_downloaded):
            self._download_file(link, file_downloaded)
                            
        if self.cohort in ['COADREAD', 'ESCA', 'GBM', 'HNSC', 'KIPAN', 'KIRC', 'KIRP', 'LAML', \
                           'LIHC', 'OV', 'PCPG', 'READ', 'SKCM', 'THYM', 'UCEC', 'UCS']:
            skiprows = 1
        else:
            skiprows = None
        self.mRNAseq = pd.read_csv(file_downloaded, header=0, skiprows=skiprows, index_col=0, sep='\t', low_memory=False)
        self.mRNAseq = self.mRNAseq.loc[['|' in i for i in self.mRNAseq.index.values.astype(str)]] # keep only gene rows
        self.mRNAseq = self.mRNAseq.astype(float)
        self.mRNAseq.index = [re.split(r"\b\|\b", idx, 1)[0] for idx in self.mRNAseq.index.values.astype(str)]
        print('Done.')
#        os.remove(file_downloaded)
        return self.mRNAseq
        
    def get_clinical(self):
        print('Retrieve clinical from http://firebrowse.org/ ...')
        print('Cohort: %s (%s)' % (self.cohortdict[self.cohort], self.cohort))
        print('File type: Clinical_Pick_Tier1')

        link = 'https://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/' + \
                self.cohort + '/20160128/gdac.broadinstitute.org_' + self.cohort + \
                '.Clinical_Pick_Tier1.Level_4.2016012800.0.0.tar.gz'

        file_downloaded = '/tmp/' + self.cohort + '.clinical.tar.gz'
        if not os.path.exists(file_downloaded):
            self._download_file(link, file_downloaded)
        with tarfile.open(file_downloaded) as f:
            clinical_file = f.extractfile('gdac.broadinstitute.org_'+self.cohort+'.Clinical_Pick_Tier1.Level_4.2016012800.0.0/' + self.cohort + '.clin.merged.picked.txt')
            self.clinical = pd.read_csv(clinical_file, header=0, index_col=0, sep='\t', low_memory=False).T
        self.clinical.index = [v.upper() for v in self.clinical.index.astype(str)]
        print('Patient barcode unique?', len(self.clinical.index) == len(np.unique(self.clinical.index)))
        print('Done.')
#        os.remove(file_downloaded)
        return self.clinical
    
    def _get_overall_survival(self, death_censor = True):
        if self.clinical is None:
            _ = self.get_clinical()
            
        colname = self.clinical.columns.values.astype(str)
        
        days_to_last_followup = self.clinical[colname[[i for i, d in enumerate(colname) if 'days_to_last_followup' in d]]].values.astype(str)
        days_to_death = self.clinical[colname[[i for i, d in enumerate(colname) if 'days_to_death' in d]]].values.astype(str)
        vital_status = self.clinical[colname[[i for i, d in enumerate(colname) if 'vital_status' in d]]].values.astype(str)
        
        print('Censorship: Dead = 1')
        e = vital_status.reshape(-1)
        e[[v in ['nan','NaN'] for v in vital_status]] = np.nan
        e = e.astype(int)
        if not death_censor:
            e = 1-e
        
        t = days_to_last_followup
        t[np.where(e == 1)] = days_to_death[np.where(e == 1)]
        t[[v in ['nan','NaN'] for v in t]] = np.nan
        t = t.reshape(-1).astype(float)
        
        return t, e
            
    def _download_file(self, link, filename):
        if not os.path.exists(filename):
            with open(filename, "wb") as f:
#                    print("Downloading %s to %s" % (self.cohort, filename))
                    response = requests.get(link, stream=True)
                    total_length = response.headers.get('content-length')
                    if total_length is None: # no content length header
                        f.write(response.content)
                    else:
                        dl = 0
                        total_length = int(total_length)
                        for data in response.iter_content(chunk_size=4096):
                            dl += len(data)
                            f.write(data)
                            done = int(50 * dl / total_length)
                            sys.stdout.write("\r[%s%s]" % ('=' * done + '>', ' ' * (50-done)) )    
                            sys.stdout.flush()
